#include "exfatlib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
   Check if the partition on lbaadr is EXFAT partition
   fat_int32 lbaadr - LBA address of the first sector
   
   Returns 0 if it is not EXFAT partition
           1 if it is     EXFAT partition
 */
int ifexfat(int drvnum, fat_int32 lbaadr)
{
  fat_int8 buf[4096];
  readsect(drvnum, lbaadr, buf);
  buf[8] = 0;
  if (!strcmp(buf+3, "EXFAT"))
    return 1;
  else
    return 0;
}

/* 
  Finds exFAT partitions on the drive
  int drvnum          - number of the physical drive
  fat_int32 *lbaadr_v - pointer to the array from 10 elements
  fat_int32 *psize_v  - pointer to the array from 10 elements

 */
int find_exfat_partitions(int drvnum, fat_int32 *lbaadr_v, fat_int32 *psize_v)
{
  int i;
  int num = 0; // Number of exFAT partitions
  fat_int8 parttype;
  fat_int32 lbaadr, psize,extended_base_addr = 0,sector_offset; 
  fat_int8 *dp; // Pointer to the partition descriptors
  fat_int8 buf[4096];
  fat_int32 next_sector_offset = 0;

  // Read MBR sector

  sector_offset = 0;

retry_extended_partitiontable:

  readsect(drvnum, sector_offset, buf);

  if (extended_base_addr == 0)
	printf("MBR partition table\n");
  else
    printf("DOS extended partition table at %X\n",sector_offset);

  // Check MBR signature
	if (getint16(buf+0x1FE) != 0xAA55)
		{
		printf("illegal partition sector at 0x%lx\n", sector_offset);
		return 1; // Invalid MBR signature - return 1
		}


  // Get the list of all partitions on the drive
	for (i = 0; i < 4; i++)
		{
		dp = buf + 0x1BE + 0x10*i; // Pointer to the i-th parition MBR record
		parttype = dp[4];

		if (parttype == 0)			// empty partition
			continue;

		lbaadr = getint32(dp+8);
		psize  = getint32(dp+12);
		printf("%d: %.2X %.2X %.8X %dMb\n", i, dp[0], parttype, lbaadr, psize/2048);

		// ? exFAT partition
		if (parttype == PART_EXFAT && ifexfat(drvnum, sector_offset+lbaadr))
			{
			printf("ExFAT partition at %X, size %X\n", lbaadr+sector_offset, psize/2048);

			lbaadr_v[num]  = lbaadr+sector_offset;
			psize_v[num++] = psize; 
			}

	    if (parttype == PART_LOGIC || parttype == PART_LOGICOLD)
			{
			next_sector_offset = extended_base_addr + lbaadr;

			if (extended_base_addr == 0)
				{
				extended_base_addr = lbaadr;
				}
			}

	if (next_sector_offset != 0)
		{
		sector_offset = next_sector_offset;
		next_sector_offset = 0;

		goto retry_extended_partitiontable;
		}
		}

  // Complete the work
  lbaadr_v[num] = 0; // Write 0 to the end of array
  psize_v[num]  = 0;                    
  return 0; // Success
}

int exfat_mount(int drvnum, fat_int32 lbaadr, exfat_descriptor *fd)
{ static char buf[4096];
  // Read VBR  
  fd->drvnum = drvnum;
  fd->lbaadr = lbaadr;
  readsect(fd->drvnum, fd->lbaadr, buf); // Load first sector of the partition
  parsevbr(&(fd->vbr), buf);
  // Fill non-VBR fields
  fd->spc = 1 << fd->vbr.sectperclust;
  fd->bps = 1 << fd->vbr.bytepersect;
  // Allocate buffers
  fd->fatbuf  = (char *) malloc(fd->bps);
  fd->fat_cs  = fd->lbaadr+fd->vbr.fatoffset;
  readsect(fd->drvnum, fd->fat_cs, fd->fatbuf); // Read first sector of the FAT
  fd->filebuf = (char *) malloc(fd->bps);
  fd->file_cs = fd->lbaadr+fd->vbr.heapoffset+fd->spc*(fd->vbr.rootclust-2);
  readsect(fd->drvnum, // Read first sector of the root directory
           fd->file_cs, 
           fd->filebuf); 
  // Return
  return 0; // Success
}

/*
    Unmount exFAT partition 
 */
int exfat_unmount(exfat_descriptor *exfat)
{
  free(exfat->fatbuf);
  free(exfat->filebuf);
  return 0;
}

/*
   Get the 32-bit exFAT entry
   exfat_descriptor *fd - descriptor of exFAT partition
   fat_int32 ind        - exFAT entry index
 */
fat_int32 getexfatentry(exfat_descriptor *fd, fat_int32 ind)
{
  fat_int64 sect;
  fat_int32 entry;
  // Calculate the sector number of FAT entry
  sect = fd->vbr.partoffset + 
         fd->vbr.fatoffset + 
         (ind >> (fd->vbr.bytepersect-2));
  // Check if the sector with the FAT entry is in the buffer
  if (sect != fd->fat_cs)
    readsect(fd->drvnum, sect, fd->fatbuf); // And read it if required
  // Return FAT entry
  entry = getint32(fd->fatbuf + (ind % (fd->bps/4))*4);
  return entry;
}

/* 
   Reads the sector of the file  
 */
int exfat_readfilesector(exfat_fd *fd, fat_int64 ind, char *buf)
{ fat_int32 cl_ind, cl_off;
  fat_int64 sectnum;
  int i;

  // Calculate the number of cluster which contains the sector
  cl_ind = (fat_int32)(ind >> fd->pd->vbr.sectperclust);
  cl_off = (fat_int32)(ind % fd->pd->spc);
  // Find absolute cluster number
  if (fd->chainflag)
  { // a) Fragmented file
    // Change the current cluster to required
    if (fd->cc_num > cl_ind)
    {
      fd->cc_ind = fd->firstclust;
      fd->cc_num = 0;
    }
    while (fd->cc_num < cl_ind)
    {
      fd->cc_ind = getexfatentry(fd->pd, fd->cc_ind);
      fd->cc_num++;
    }
  }
  else
  { // b) not fragmented file
    fd->cc_ind = fd->firstclust + cl_ind;
    fd->cc_num = cl_ind;
  }

  // Calculate the absolute sector index
  sectnum = fd->pd->lbaadr + 
            fd->pd->vbr.heapoffset + 
            fd->pd->spc*(fd->cc_ind-2) + cl_off;
  // Read the sector into the filebuffer
  if (sectnum != fd->pd->file_cs)
  {
    readsect(fd->pd->drvnum, sectnum, fd->pd->filebuf);
    fd->pd->file_cs = (__uint32)sectnum;
  }
  // Copy the result in the buffer of user
  if (buf != NULL)
  {
    for (i = 0; i < fd->pd->bps; i++)
      buf[i] = fd->pd->filebuf[i];
  }
//  printf("<<%d %d>>\n", cl_ind, cl_off);   

  return 0; //??
}

/*
   Reads folder entry in fd->pos position
   (function for internal use only)
   exfat_fd *fd - pointer to the file desriptor of the folder
   
 */
char *exfat_getfolderentry(exfat_fd *fd)
{
  fat_int64 ind; // Sector number with the folder entry
  fat_int32 adr; // Offset into the sector in bytes
  // Calculate the sector and offset for the current
  // position in the file
  ind = (__uint32)(fd->pos >> (fd->pd->vbr.bytepersect)); // Sector
  adr = (__uint32)(fd->pos %  (fd->pd->bps));             // Offset
  // Read sector  
  exfat_readfilesector(fd, ind, NULL);
  // Return the beginning of folder entry in the buffer
  return fd->pd->filebuf + adr;
}

/* 
   Finds next entry in the folder 
   exfat_fd *fd - pointer to the file desriptor of the folder
   exfat_fd *outfd - pointer to the file descriptor for output
*/
int exfat_findnext(exfat_fd *fd, exfat_fd *outfd)
{ char *ep;
  int i,j,n;

  int namelen; // Length of the file name
  fat_int16 *fnamebuf; // Buffer for UNICODE file name
  // Go the beginning of the folder
  do 
  {
    ep = exfat_getfolderentry(fd);
    fd->pos += 32; // To the next entry
  } while(*ep != DE_END && *ep != DE_FILE);

  if (*ep == DE_END)
    return 0; // Empty entry
  else
  { // Get file attributes
    outfd->attrib = getint16(ep+4);
    // Process not empty entry
    // a) Go to the stream extension directory entry
    // to get name length, first cluster, and file length
    ep = exfat_getfolderentry(fd);
    fd->pos += 32;
    if (*ep == DE_STREAM)
    {
      namelen           = ep[3];
      outfd->firstclust = getint32(ep+20);
      outfd->filesize   = getint64(ep+24);
      outfd->chainflag  = (ep[1] & 0x02) == 0; 
      outfd->pd         = fd->pd;
      outfd->pos        = 0;
      outfd->cc_num     = 0;
      outfd->cc_ind     = outfd->firstclust;
    }
    else
    { // !!! FIXME process error
    } 
    // b) read the file name
    // Allocate buffer
    if (outfd->filename != NULL)
      free(outfd->filename); 
    outfd->filename = (fat_int16 *) calloc(namelen+1, 2);
    fnamebuf = outfd->filename;
    i = namelen; n = 0;
    while(i > 0)
    { i-=15; n++; }
    for (i = 0; i < n; i++)
    {
      ep = exfat_getfolderentry(fd);
      fd->pos += 32; 
      if (*ep == DE_FNAME)
      {
        for (j = 0; j < 15; j++)
        { if (i*15+j >= namelen) 
          { fnamebuf[i*15+j] = 0; 
            break;
          }
          fnamebuf[i*15+j] = getint16(ep+(j+1)*2);
        }
      }
      else
      { // !!! FIXME process error
        printf("<!>");  
      }
    }
    return 1;
  }
}

/* 
 */

int exfat_findfirst(exfat_fd *fd, exfat_fd *outfd)
{
  fd->pos = 0;
  return exfat_findnext(fd, outfd);
}

/* 
   Finds next entry in the folder 
   fd       - pointer to the file descriptor of the folder (directory)
   outfd    - pointer to the output file descriptor 
   filename - pointer to the UNICODE \0-terminating string with the file name
 
   0 - success
   1 - not found
 */

int exfat_findfile(exfat_fd *fd, exfat_fd *outfd, fat_int16 *filename)
{
  exfat_findfirst(fd, outfd);
  do {
    if (unicode_strcmp(filename, outfd->filename) == 0)
    {
      return 0;
    }
  } while (exfat_findnext(fd, outfd));
  return 1;
}

/*
  Parse the path and get file descriptor
  path - path in UNICODE \0-terminated string
  fd_root - pointer to the descriptor of root directory
  fd      - pointer to the descriptor for the result
 */
int exfat_parsepath(fat_int16 *path, exfat_fd *fd_root, exfat_fd *fd)
{
  fat_int16 fname[256]; // File name between backslashes
  int i, iprev, j;
  exfat_fd fdprev;

  *fd    = *fd_root;
  fdprev = *fd_root; i = 0;
  do 
  { // Take file name between backslashes
    iprev = i;
    while (path[i] != '\\' && path[i] != '\0') i++;
    i++;
    for (j = iprev; j < i; j++)
      fname[j-iprev] = path[j];
    fname[i-iprev] = 0;
    if (fname[i-iprev-1] == '\\')
      fname[i-iprev-1] = 0;
    // Exclude empty name and name of root folders
    if (fname[2] == ':' || fname[0] == 0) continue;
    // Process file name
    if (exfat_findfile(&fdprev, fd, fname) == 0)
    {
      // !!! FIXME ERASE OLD FD DESCRIPTOR
      fdprev = *fd;
    }
    else
    {
      return 1;
    }
  } while (path[i-1] != 0);
  return 0;
}
